//*****************************************************************************
//
//  File........: Temperature.c
//
//  Description.: Temperature measurements
//
//*****************************************************************************

//  Include files

#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <Temperature.h>
#include <math.h>



extern long int set_point;
extern float TEMP_HIGHBYTE, TEMP_LOWBYTE;
//  Global variables

unsigned char ADCresult_calc_cnt = 0;   // variable to count number of temperature calcualtions
int ADCresult_average = 0;              // integer to keep the ADC results for later averaging
float Temperature;                      // float to calculate the temperature

/******************************************************************************************
*
*	Function name : ADC_init
*
*	Returns :		None
*
*	Parameters :	Operation, selects between Single Ended or Differential ADC
*
*	Purpose :		Initialize the ADC and run a dummy ADC conversion
*
******************************************************************************************/
void ADC_init(unsigned char Operation)
{
    set_point = 25;                     // set SET_POINT to 25 degree Celsius

    ADMUX |= Operation;                 // set the selected operation, either Single Ended or Differential
        
  //  Delay(50);                          // delay in order for channel selection to settle 
    
    ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS0);    // set ADC prescaler to , 1MHz / 8 = 125kHz

  //  DIDR1 = 0xFF;                       // disable digital input on PORTF
    
    ADC_conversion();                   // run a dummy conversion the first time the ADC is used after a power up
}

/******************************************************************************************
*
*	Function name : PWM_init
*
*	Returns :		None
*
*	Parameters :	None
*
*	Purpose :		Initialize the Timer1 with PWM, Phase and Frequency Correct, 10-bit
*
******************************************************************************************/
void PWM_init(void)
{
    ICR1H = 0x03;               // set the TOP value for the PWM to 0x3FF = 10bit
    ICR1L = 0xFF;
    
    OCR1AH = OCR1BH = 0x00;
    OCR1AL = OCR1BL = 0xF0;     // set a initial value in the OCR1A-register
   
    TCCR1A = (1<<COM1A1) | (1<<COM1A0) | (1<<COM1B1) | (1<<COM1B0); // set OC1A/B on upcounting, clear on downcounting
    TCCR1B = (1<<WGM13) | (1<<CS11);        // select Phase and Frequency Corrcet, and set CLK(I/O) / 8 (From prescaler)
}

/******************************************************************************************
*
*	Function name : ADC_conversion
*
*	Returns :		None
*
*	Parameters :	None
*
*	Purpose :		Run the ADC conversion 
*
******************************************************************************************/
void ADC_conversion(void)
{
    int ADC_temp;
    int ADCresult = 0;
    unsigned char i;
    
    for(i=0;i<8;i++)            // do the ADC conversion 8 times for better accuracy 
    {
        ADCSRA |= (1<<ADSC);        // do single conversion
        while(!(ADCSRA & 0x10));    // wait for conversion done, ADIF flag active

        ADC_temp = ADCL;            // read out ADCL register
        ADC_temp += (ADCH << 8);    // read out ADCH register        

        ADCresult += ADC_temp;      // accumulate result (8 samples) for later averaging
    }
    
    ADCresult = ADCresult >> 3;     // average the 8 samples

    ADCresult_calc_cnt++;           // increment the counter that keeps track of how many calulations that have been donet 
    
    ADCresult_average += ADCresult; // add the ADCresult to the previous measurements
    
    if(ADCresult_calc_cnt >= 32)    // if 32 calculations
    {
        ADCresult_calc_cnt = 0;     // clear the counter
       
        ADCresult_average = ADCresult_average >> 5;   // find the average ADC result for the last 32 times
        
        Temp_calculation(ADCresult_average);       // call the temperature calculation function

        ADCresult_average = 0;      // clear the ADCresult_average
    }  


//*****************************************************
//Temperature = 35;
//*****************************************************


}

/******************************************************************************************
*
*	Function name : Temp_calculation
*
*	Returns :		None
*
*	Parameters :	The result in the ADCH/L registers
*
*	Purpose :		Calculate the corresponding temperature in degree Celsius, 
*                   and place the value in SRAM at location 0x4F0 - 0x4FF
*
******************************************************************************************/
void Temp_calculation(int iADC_value)
{
    float V_ADC;
    
    if((ADMUX & 0x1F) == SingleEnded)       // check if Single_ended operation
        V_ADC = (iADC_value * V_ref)/1024;      // calculate the voltage over the NTC
    else if((ADMUX & 0x1F) == Differential) // check if Differential operation
        V_ADC = (iADC_value * V_ref)/512;       // calculate the voltage over the NTC

    Temperature = (Beta / ((log(V_ADC/(V_ref-V_ADC))/log(__E)) + (Beta/T_amb))) - T_zero;   // calculate the temperature
    
    TEMP_HIGHBYTE = Temperature;                // store the integer to "TEMP_HIGHBYTE"
    //Temperature = (Temperature_average - TEMP_HIGHBYTE);    // subtract the integer from Temperature_average        
    Temperature = (1.8 * Temperature) + 32;
	TEMP_LOWBYTE = (Temperature * 100);         // multiply Temperature_average by 100 to get the decimal, and store it to "TEMP_LOWBYTE"

    //Temperature_regulation();                   //call the Temperatur_regulation rutine
}

/******************************************************************************************
*
*	Function name : Temperature_regulation
*
*	Returns :	    None    	
*
*	Parameters :	None
*
*	Purpose :		Find the difference between measured and requested temperature. And set 
*                   a proper value in the OCR1A-register for the use of PWM on the Leds.
*
******************************************************************************************/
void Temperature_regulation(void)
{
    
}
